/*
 * SMBlib NetBIOS implementation
 * 
 * Version 1.0
 * SMBlib Routines. Logon (experimental) Section ...
 * 
 * Copyright (C) Richard Sharpe 1996
 * 
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"


/*
 * Logon and tree connect to the server. If a tree handle was given to us,
 * we use it and return it, otherwise we create one ...
 */
struct SMB_Tree *SMB_Logon_And_TCon (struct SMB_Connect *connHandle,
                                     struct SMB_Tree    *treeHandle,
                                     const  char        *userName,
                                     const  char        *passWord,
                                     const  char        *service,
                                     const  char        *service_type)
{
  struct RFCNB_Pkt *pkt;
  struct SMB_Tree  *tree;
  int    param_len, pkt_len, andx_len, andx_param_len;
  char  *p, *AndXCom;

  /* Lets create a tree if we need one ...
   */
  if (!treeHandle)
  {
    tree = calloc (sizeof(*tree), 1);
    if (!tree)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (tree);
    }
    /* Initialize the tree
     */
    tree->con = connHandle;
    tree->prev = tree->next = NULL;
  }
  else
    tree = treeHandle;

  /* First we need a packet etc ... but we need to know what protocol has
   * been negotiated to figure out if we can do it and what SMB format to
   * use ...
   */

  /* Since we are going to do a LogonAndX with a TCon as the second command
   * We need the packet size correct. So TCon starts at wct field
   */
  if (connHandle->protocol < SMB_P_LanMan1)
  {
    SMBlib_errno = SMBlibE_ProtLow;
    if (treeHandle == NULL)
       free (tree);
    return (NULL);
  }

  /* Now build the correct structure
   */
  andx_len = SMB_tconx_len - SMB_hdr_wct_offset;

  /* We send a null password as we sent one in the setup and X
   */
  andx_param_len = strlen (service) + 1 + strlen (service_type) + 1;

  if (connHandle->protocol < SMB_P_NT1)
  {
    SMB_debug (1, "Doing an LM session setup etc ...\n");

    /* We don't do encrypted passwords ...
     */
    param_len = strlen (userName) + 1 + strlen (passWord) + 1 +
                strlen (connHandle->PDomain) + 1 +
                strlen (connHandle->OSName) + 1;

    pkt_len = SMB_ssetpLM_len + param_len + andx_len + andx_param_len;
    pkt     = RFCNB_Alloc_Pkt (pkt_len);
    if (!pkt)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      if (!treeHandle)
         free (tree);
      return (NULL);
    }

    bzero (SMB_Hdr (pkt), SMB_ssetpLM_len);
    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, connHandle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, connHandle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, 0);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 10;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = SMBtconX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, SMB_ssetpLM_len + param_len);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mmc_offset, 2);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_vcn_offset, connHandle->pid);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_pwl_offset, strlen (passWord) + 1);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_res_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char *) (SMB_Hdr (pkt) + SMB_ssetpLM_buf_offset);

    /* Copy  in password, then the rest. Password has no null at end
     */
    strcpy (p, passWord);
    p += strlen (passWord) + 1;

    strcpy (p, userName);
    p += strlen (userName);
    *p++ = '\0';

    strcpy (p, connHandle->domain);
    p += strlen (connHandle->PDomain);
    *p++ = '\0';

    strcpy (p, connHandle->OSName);
    p += strlen (connHandle->OSName);
    *p = '\0';

    AndXCom = SMB_Hdr (pkt) + SMB_ssetpLM_len + param_len - SMB_hdr_wct_offset;
  }
  else
  {
    /* We don't admit to UNICODE support ...
     */
    SMB_debug (1, "Doing NT LM Sess Setup etc ... \n");

    param_len = strlen (userName) + 1 + strlen (passWord) +
                strlen (connHandle->PDomain) + 1 +
                strlen (connHandle->OSName) + 1 +
                strlen (connHandle->LMType) + 1;

    pkt_len = SMB_ssetpNTLM_len + param_len + andx_len + andx_param_len;
    pkt     = RFCNB_Alloc_Pkt (pkt_len);
    if (!pkt)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      if (!treeHandle)
         free (tree);
      return (NULL);
    }

    bzero (SMB_Hdr (pkt), SMB_ssetpNTLM_len);
    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, connHandle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, connHandle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, 0);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 13;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = SMBtconX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, SMB_ssetpNTLM_len + param_len);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mmc_offset, 2);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_vcn_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cipl_offset, strlen (passWord));
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cspl_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_res_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cap_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char *) (SMB_Hdr (pkt) + SMB_ssetpNTLM_buf_offset);

    /* Copy  in password, then the rest. Password has no null at end
     */
    strcpy (p, passWord);
    p += strlen (passWord);

    strcpy (p, userName);
    p += strlen (userName);
    *p++ = '\0';

    strcpy (p, connHandle->PDomain);
    p += strlen (connHandle->PDomain);
    *p++ = '\0';

    strcpy (p, connHandle->OSName);
    p += strlen (connHandle->OSName);
    *p++ = '\0';

    strcpy (p, connHandle->LMType);
    p += strlen (connHandle->LMType);
    *p = '\0';

    /* Now set up the TCON Part ... from WCT, make up a pointer that will
     * help us ...
     */
    AndXCom = SMB_Hdr(pkt) + SMB_ssetpNTLM_len + param_len -
              SMB_hdr_wct_offset;
  }
  *(AndXCom + SMB_hdr_wct_offset) = 4;
  *(AndXCom + SMB_tconx_axc_offset) = 0xFF;   /* No command */
  SSVAL (AndXCom, SMB_tconx_axo_offset, 0);
  SSVAL (AndXCom, SMB_tconx_flg_offset, 0);   /* Don't disconnect TID */
  SSVAL (AndXCom, SMB_tconx_pwl_offset, 0);   /* No password, */
  SSVAL (AndXCom, SMB_tconx_bcc_offset, andx_param_len);

  p = (char *) (AndXCom + SMB_tconx_buf_offset);
  strcpy (p, service);
  p += strlen (service) + 1;
  strcpy (p, service_type);

  /* Now send it and get a response
   */
  if (RFCNB_Send (connHandle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error sending SessSetupAndTCon request\n");

    RFCNB_Free_Pkt (pkt);
    free (tree);
    SMBlib_errno = SMBlibE_SendFailed;
    return (NULL);
  }

  /* Now get the response ...
   */
  if (RFCNB_Recv (connHandle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error receiving response to SessSetupAndTCon\n");

    RFCNB_Free_Pkt (pkt);
    free (tree);
    SMBlib_errno = SMBlibE_RecvFailed;
    return (NULL);
  }

  /* Check out the response type ...
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_debug (1, "SMB_SessSetupAndTCon fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset));

    /* Note, here, that we have not properly handled the error processing
     * and so we cannot tell how much of our request crapped out
     */
    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    free (tree);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (NULL);
  }

  SMB_debug (1, "SessSetupAndX response. Action = %i\n",
             SVAL (SMB_Hdr (pkt), SMB_ssetpr_act_offset));

  /* Now pick up the UID for future reference ...
   */
  connHandle->uid = SVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset);

  /* And pick up the TID as well
   */
  tree->tid = SVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset);

  tree->mbs = connHandle->max_xmit;

  /* Link the tree into the list in con
   */
  if (connHandle->first_tree == NULL)
  {
    connHandle->first_tree = tree;
    connHandle->last_tree  = tree;
  }
  else
  {
    connHandle->last_tree->next = tree;
    tree->prev = connHandle->last_tree;
    connHandle->last_tree = tree;
  }

  RFCNB_Free_Pkt (pkt);

  return (tree);
}

/*
 * Logon and TCon and Open to a file on the server, but we need to pass
 * back a file pointer, so we better have one in the parameter list
 */
int SMB_Logon_TCon_Open (struct SMB_Connect *connHandle,
                         const  char        *userName,
                         const  char        *passWord,
                         const  char        *service,
                         const  char        *service_type,
                         struct SMB_Tree   **treeHandle,
                         const  char        *filename,
                         WORD                mode,
                         WORD                search,
                         struct SMB_File   **File_Handle)
{
  struct RFCNB_Pkt  *pkt;
  struct SMB_File   *file_tmp;
  struct SMB_Tree   *tree;
  int    param_len, pkt_len, tcon_len, tcon_param_len;
  int    open_len, open_param_len, header_len;
  char  *p, *AndXCom;

  /* First, we need a tree STRUCTURE as we are going to tree connect
   */
  tree = calloc (sizeof(*tree), 1);
  if (!tree)
  {
    SMBlib_errno = SMBlibE_NoSpace;
    return (SMBlibE_BAD);
  }
  tree->con  = connHandle;
  tree->next = tree->prev = NULL;

  /* Next, we need a file handle as we are going to pass one back ...
   * Hmm, there is a bug here ... We should check on File_Handle ...
   */
  file_tmp = calloc (sizeof(*file_tmp), 1);
  if (!file_tmp)
  {
    SMB_debug (0, "Could not allocate file handle space ...");
    SMBlib_errno = SMBlibE_NoSpace;
    free (tree);
    return (SMBlibE_BAD);
  }

  /* Next we need a packet etc ... but we need to know what protocol has
   * been negotiated to figure out if we can do it and what SMB format to
   * use ...
   */

  /* Since we are going to do a LogonAndX with a TCon as the second command
   * We need the packet size correct. So TCon starts at wct field
   */
  if (connHandle->protocol < SMB_P_LanMan1)
  {
    free (tree);
    free (file_tmp);
    SMBlib_errno = SMBlibE_ProtLow;
    return (SMBlibE_BAD);
  }

  /* Now build the correct structure
   *
   * We send a null password in the TconAndX ...
   */
  tcon_len       = SMB_tconx_len - SMB_hdr_wct_offset;
  tcon_param_len = strlen (service) + 1 + strlen (service_type) + 1;

  open_len       = SMB_openx_len - SMB_hdr_wct_offset;
  open_param_len = 1 + strlen (filename) + 1; /* AsciiID + null */

  if (connHandle->protocol < SMB_P_NT1)
  {
    /* We don't do encrypted passwords yet
     */
    param_len = strlen (userName) + 1 + strlen (passWord) + 1 +
                strlen (connHandle->PDomain) + 1 +
                strlen (connHandle->OSName) + 1;

    header_len = SMB_ssetpLM_len + param_len;
    pkt_len    = header_len + tcon_len + tcon_param_len +
                 open_len + open_param_len;
    pkt        = RFCNB_Alloc_Pkt (pkt_len);
    if (!pkt)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      free (tree);
      free (file_tmp);
      return (SMBlibE_BAD);     /* Should handle the error */
    }

    bzero (SMB_Hdr (pkt), SMB_ssetpLM_len);
    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, connHandle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, connHandle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, 0);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 10;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = SMBtconX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, SMB_ssetpLM_len + param_len);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mmc_offset, 2);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_vcn_offset, connHandle->pid);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_pwl_offset, strlen (passWord) + 1);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_res_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char *) (SMB_Hdr (pkt) + SMB_ssetpLM_buf_offset);

    /* Copy  in password, then the rest. Password has no null at end
     */
    strcpy (p, passWord);
    p += strlen (passWord) + 1;

    strcpy (p, userName);
    p += strlen (userName);
    *p++ = '\0';

    strcpy (p, connHandle->PDomain);
    p += strlen (connHandle->PDomain);
    *p++ = '\0';

    strcpy (p, connHandle->OSName);
    p += strlen (connHandle->OSName);
    *p = '\0';

    AndXCom = SMB_Hdr (pkt) + SMB_ssetpLM_len + param_len -
              SMB_hdr_wct_offset;
  }
  else
  {
    /* We don't admit to UNICODE support ...
     */
    param_len = strlen (userName) + 1 + strlen (passWord) +
                strlen (connHandle->PDomain) + 1 +
                strlen (connHandle->OSName) + 1 +
                strlen (connHandle->LMType) + 1;

    header_len = SMB_ssetpNTLM_len + param_len;
    pkt_len    = header_len + tcon_len + tcon_param_len +
                 open_len + open_param_len;
    pkt        = RFCNB_Alloc_Pkt (pkt_len);
    if (!pkt)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      free (tree);
      free (file_tmp);
      return (-1);
    }

    bzero (SMB_Hdr (pkt), SMB_ssetpNTLM_len);
    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, connHandle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, connHandle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, 0);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 13;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = SMBtconX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, SMB_ssetpNTLM_len + param_len);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mmc_offset, 2);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_vcn_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cipl_offset, strlen (passWord));
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cspl_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_res_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cap_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char *) (SMB_Hdr (pkt) + SMB_ssetpNTLM_buf_offset);

    /* Copy  in password, then the rest. Password has no null at end
     */
    strcpy (p, passWord);
    p += strlen (passWord);

    strcpy (p, userName);
    p += strlen (userName);
    *p++ = '\0';

    strcpy (p, connHandle->PDomain);
    p += strlen (connHandle->PDomain);
    *p++ = '\0';

    strcpy (p, connHandle->OSName);
    p += strlen (connHandle->OSName);
    *p++ = '\0';

    strcpy (p, connHandle->LMType);
    p += strlen (connHandle->LMType);
    *p = '\0';

    /* Now set up the TCON Part ... from WCT, make up a pointer that will
     * help us ...
     */
    AndXCom = SMB_Hdr (pkt) + SMB_ssetpNTLM_len + param_len -
              SMB_hdr_wct_offset;
  }

  *(AndXCom + SMB_hdr_wct_offset)   = 4;
  *(AndXCom + SMB_tconx_axc_offset) = SMBopenX;
  SSVAL (AndXCom, SMB_tconx_axo_offset, (header_len + tcon_len + tcon_param_len));
  SSVAL (AndXCom, SMB_tconx_flg_offset, 0); /* Don't disconnect TID    */
  SSVAL (AndXCom, SMB_tconx_pwl_offset, 0); /* No password */
  SSVAL (AndXCom, SMB_tconx_bcc_offset, tcon_param_len);

  p = (char *) (AndXCom + SMB_tconx_buf_offset);
  strcpy (p, service);
  p += strlen (service) + 1;
  strcpy (p, service_type);

  /* Now the open bit ...
   */
  AndXCom = AndXCom + tcon_len + tcon_param_len; /* Should get us there */

  *(AndXCom + SMB_hdr_wct_offset) = 15;
  *(AndXCom + SMB_openx_axc_offset) = 0xFF;
  *(AndXCom + SMB_openx_axr_offset) = 0;
  SSVAL (AndXCom, SMB_openx_axo_offset, 0);
  SSVAL (AndXCom, SMB_openx_flg_offset, 0);
  SSVAL (AndXCom, SMB_openx_mod_offset, mode);
  SSVAL (AndXCom, SMB_openx_atr_offset, search);
  SSVAL (AndXCom, SMB_openx_fat_offset, 0);
  SIVAL (AndXCom, SMB_openx_tim_offset, 0);
  SSVAL (AndXCom, SMB_openx_ofn_offset, 0x0011); /* Create or open */
  SIVAL (AndXCom, SMB_openx_als_offset, 0);
  SSVAL (AndXCom, SMB_openx_bcc_offset, open_param_len);

  p = (char*) (AndXCom + SMB_openx_buf_offset);
  strcpy (p, filename);

  /* Now send it and get a response
   */
  if (RFCNB_Send (connHandle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error sending SessSetupAndTCon request\n");

    RFCNB_Free_Pkt (pkt);
    free (tree);
    free (file_tmp);
    SMBlib_errno = SMBlibE_SendFailed;
    return (SMBlibE_BAD);
  }

  /* Now get the response ...
   */
  if (RFCNB_Recv (connHandle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error receiving response to SessSetupAndTCon\n");

    RFCNB_Free_Pkt (pkt);
    free (tree);
    free (file_tmp);
    SMBlib_errno = SMBlibE_RecvFailed;
    return (SMBlibE_BAD);
  }

  /* Check out the response type ...
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {
    SMB_debug (1, "SMB_SessSetupAndTCon fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset));

    /* Note, here, that we have not properly handled the error processing
     * and so we cannot tell how much of our request crapped out
     */
    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    RFCNB_Free_Pkt (pkt);
    free (tree);
    free (file_tmp);
    SMBlib_errno = SMBlibE_Remote;
    return (SMBlibE_BAD);
  }

  SMB_debug (1, "SessSetupAndX response. Action = %i\n",
             SVAL (SMB_Hdr (pkt), SMB_ssetpr_act_offset));

  /* Now pick up the UID for future reference ...
   */
  connHandle->uid = SVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset);

  /* And pick up the TID as well
   */
  tree->tid = SVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset);
  tree->mbs = connHandle->max_xmit; /* We need this */

  SMB_debug (2, "mbs=%lu\n", tree->mbs);

  /* Now we populate the file hanble and pass it back ...
   */
  strncpy (file_tmp->filename, filename, sizeof (file_tmp->filename) - 1);
  file_tmp->tree = tree;

  /* Pick up a pointer to the right part ...
   */
  AndXCom = SMB_Hdr (pkt) + SVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset) -
            SMB_hdr_wct_offset;

  /* Now skip the response to the TConX
   */
  AndXCom = SMB_Hdr (pkt) + SVAL (AndXCom, SMB_tconxr_axo_offset) -
            SMB_hdr_wct_offset;

  SMB_debug (2, "Word Params = %x, AXO = %x\n",
             CVAL (AndXCom, SMB_hdr_wct_offset),
             SVAL (AndXCom, SMB_openxr_axo_offset));

  /* Now pick up the things from the openX response that we need
   */
  file_tmp->fid     = SVAL (AndXCom, SMB_openxr_fid_offset);
  file_tmp->lastmod = IVAL (AndXCom, SMB_openxr_tim_offset);
  file_tmp->size    = IVAL (AndXCom, SMB_openxr_fsz_offset);
  file_tmp->access  = SVAL (AndXCom, SMB_openxr_acc_offset);
  file_tmp->fileloc = 0;

  *File_Handle = file_tmp;

  /* Now link the tree into the right place ...
   */
  if (connHandle->first_tree == NULL)
  {
    connHandle->first_tree = tree;
    connHandle->last_tree  = tree;
  }
  else
  {
    connHandle->last_tree->next = tree;
    tree->prev = connHandle->last_tree;
    connHandle->last_tree = tree;
  }

  RFCNB_Free_Pkt (pkt);

  *treeHandle = tree;
  return (0);
}
