/*
 * SMBlib NetBIOS implementation
 * 
 * Version 1.0
 * SMBlib Routines
 * 
 * Copyright (C) Richard Sharpe 1996
 * 
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "std-defs.h"
#include "extract.h"
#include "rfcnb.h"
#include "smblib.h"

int SMBlib_errno     = 0;
int SMBlib_SMB_Error = 0;
int SMBlib_debug     = 0;

static char  *SMB_def_domain = SMBLIB_DEFAULT_DOMAIN;
static DWORD  SMB_def_server = 0;
static char  *SMB_def_share  = NULL;

static char  *SMB_def_myhost = NULL;
static char  *SMB_def_user   = NULL;
static char  *SMB_def_passwd = NULL;

static char  *SMB_def_osname = SMBLIB_DEFAULT_OSNAME;
static char  *SMB_def_lmType = SMBLIB_DEFAULT_LMTYPE;

/*
 * Wattcp.cfg config hook handler. Parses "SMB.xx" keywords
 */
static void (*prev_cfg_hook) (const char*, const char*) = NULL;

static void smb_config (const char *name, const char *value)
{
  static struct config_table smb_cfg[] = {
                { "DOMAIN",   ARG_STRDUP,  (void*)&SMB_def_domain },
                { "SERVER",   ARG_RESOLVE, (void*)&SMB_def_server },
                { "SHARE",    ARG_STRDUP,  (void*)&SMB_def_share  },
                { "MYHOST",   ARG_STRDUP,  (void*)&SMB_def_myhost },
                { "USERNAME", ARG_STRDUP,  (void*)&SMB_def_user   },
                { "PASSWORD", ARG_STRDUP,  (void*)&SMB_def_passwd },
                { "OSNAME",   ARG_STRDUP,  (void*)&SMB_def_osname },
                { "LMTYPE",   ARG_STRDUP,  (void*)&SMB_def_lmType },
                { "DEBUG",    ARG_ATOI,    (void*)&SMBlib_debug   },
                { NULL }
              };

  if (!parse_config_table(&smb_cfg[0], "SMB.", name, value) &&
      prev_cfg_hook)
     (*prev_cfg_hook) (name, value);
}

/*
 * Initialize the SMBlib package
 */
int SMB_Init (void)
{
  if (my_ip_addr)    /* not completely fool-proof! */
  {
    fprintf (stderr, "SMB_Init() must be called before sock_init()\n");
    return (-1);
  }

  prev_cfg_hook = usr_init;
  usr_init = smb_config;
  return (0);
}

/*
 * SMB_Create: Create a connection structure and return for later use
 * We have other helper routines to set variables
 */
struct SMB_Connect *SMB_Create_Con_Handle (void)
{
  SMBlib_errno = SMBlibE_NotImpl;
  return (NULL);
}

int SMBlib_Set_Sock_NoDelay (struct SMB_Connect *con, BOOL yn)
{
  if (RFCNB_Set_Sock_NoDelay (con->transConnect, yn) < 0)
     SMB_debug (1, "Setting no-delay on TCP socket fail: %s",
                strerror(errno));
  return (0);
}

/*
 * Get our hostname
 */
void SMB_Get_My_Name (char *name, int len)
{
  char *dot;

  if (SMB_def_myhost)
  {
    strncpy (name, SMB_def_myhost, len);
    return;
  }
  if (gethostname (name, len) < 0)
  { 
    strncpy (name, "unknown", len);
    SMB_debug (1, "SMB_Get_My_Name: %s\n", strerror(errno));
    return;
  }
  dot = strchr (name, '.');    /* extract the host-part */
  if (dot)
     *dot = '\0';
}

/*
 * SMB_Connect_Server: Connect to a server, but don't negotiate
 * protocol or anything else ...
 */
struct SMB_Connect *SMB_Connect_Server (struct SMB_Connect *handle,
                                        const  char        *server)
{
  struct SMB_Connect *con = handle;
  char   called[80], calling[80], *address;

  /* Get a connection structure if one does not exist
   */
  if (!handle)
  {
    con = calloc (sizeof(*con), 1);
    if (!con)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (NULL);
    }
  }

  /* Init some things ...
   */
  con->service [0] = '\0';
  con->username[0] = '\0';
  con->password[0] = '\0';
  con->address [0] = '\0';
  con->sock_options[0] = '\0';

  strcpy (con->desthost, server);
  strcpy (con->PDomain, SMB_def_domain);
  strcpy (con->OSName,  SMB_def_osname);
  strcpy (con->LMType,  SMB_def_lmType);
  con->first_tree = con->last_tree = NULL;

  SMB_Get_My_Name (con->myname, sizeof(con->myname));

  con->port = 0;                /* No port selected */

  /* Get some things we need for the SMB Header
   */
  con->pid = getpid();
  con->mid = con->pid;
  con->uid = 0;
  con->gid = getgid();

  /* Now connect to the remote end, but first upper case the name of the
   * service we are going to call, sine some servers want it in uppercase
   */
  strncpy (called, server, sizeof(called)-1);
  called [sizeof(called)-1] = '\0';
  strupr (called);

  strncpy (calling, con->myname, sizeof(calling)-1);
  calling [sizeof(calling)-1] = '\0';
  strupr (calling);

  address = (con->address[0] ? con->address : con->desthost);

  con->transConnect = RFCNB_Call (called, calling, address, 0);
  if (!con->transConnect)
  {
    if (!handle)
       free (con);
    SMBlib_errno = -(SMBlibE_CallFailed);
    return (NULL);
  }
  return (con);
}

/*
 * SMB_Connect: Connect to the indicated server
 * If Con_Handle == NULL then create a handle and connect, otherwise
 * use the handle passed
 */
char *SMB_Prots_Restrict[] = { "PC NETWORK PROGRAM 1.0",
                               NULL
                             };

struct SMB_Connect *SMB_Connect (struct SMB_Connect *handle,
                                 struct SMB_Tree   **tree,
                                 const  char        *service,
                                 const  char        *username,
                                 const  char        *password)
{
  struct SMB_Connect *con = handle;
  char  *host, *address;
  char   temp[80], called[80], calling[80];
  int    i;

  /* Get a connection structure if one does not exist
   */
  if (!handle)
  {
    con = calloc (sizeof(*con), 1);
    if (!con)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (NULL);
    }
  }

  /* Init some things ...
   */
  strcpy (con->service,  service);
  strcpy (con->username, username);
  strcpy (con->password, password);
  con->sock_options[0] = con->address[0] = '\0';

  strcpy (con->PDomain,SMB_def_domain);
  strcpy (con->OSName, SMB_def_osname);
  strcpy (con->LMType, SMB_def_lmType);
  con->first_tree = con->last_tree = NULL;

  SMB_Get_My_Name (con->myname, sizeof(con->myname));

  con->port = 0;                /* No port selected */

  /* Get some things we need for the SMB Header
   */
  con->pid = getpid();
  con->mid = con->pid;          /* This will do for now ... */
  con->uid = 0;                 /* Until we have done a logon, no uid */
  con->gid = getgid();

  /* Now figure out the host portion of the service
   */
  strcpy (temp, service);
  host = strtok (temp, "/\\");  /* Separate host name portion */
  strcpy (con->desthost, host);

  /* Now connect to the remote end, but first upper case the name of the
   * service we are going to call, sine some servers want it in uppercase
   */
  for (i = 0; i < strlen(host); i++)
      called[i] = toupper (host[i]);

  called [i] = '\0';

  for (i = 0; i < strlen(con->myname); i++)
      calling[i] = toupper (con->myname[i]);

  calling [i] = '\0';

  if (con->address[0] == '\0')
       address = con->desthost;
  else address = con->address;

  con->transConnect = RFCNB_Call (called, calling, address, con->port);
  if (!con->transConnect)
  {
    if (!handle)
       free (con);
    SMBlib_errno = -(SMBlibE_CallFailed);
    return (NULL);
  }

  /* Now, negotiate the protocol
   */
  if (SMB_Negotiate (con, SMB_Prots_Restrict) < 0)
  {
    /* Hmmm what should we do here ... We have a connection, but
     * could not negotiate ...
     */
    return (NULL);
  }

  /* Now connect to the service ...
   */
  *tree = SMB_TreeConnect (con, NULL, service, password, "A:");
  if (*tree == NULL)
     return (NULL);
  return (con);
}

/*
 * Logon to the server. That is, do a session setup if we can.
 * We do not do Unicode yet!
 */
int SMB_Logon_Server (struct SMB_Connect *handle,
                      const  char        *userName,
                      const  char        *passWd)
{
  struct RFCNB_Pkt *pkt;
  int    param_len, pkt_len;
  char  *p;

  /* First we need a packet etc ... but we need to know what protocol has
   * been negotiated to figure out if we can do it and what SMB format to
   * use ...
   */
  if (handle->protocol < SMB_P_LanMan1)
  {
    SMBlib_errno = SMBlibE_ProtLow;
    return (-1);
  }

  /* Now build the correct structure
   */
  if (handle->protocol < SMB_P_NT1)
  {
    /* We don't handle encrypted passwords ...
     */
    param_len = strlen (userName) + 1 +
                strlen (passWd) + 1 +
                strlen (handle->PDomain) + 1 +
                strlen (handle->OSName) + 1;

    pkt_len = SMB_ssetpLM_len + param_len;
    pkt     = RFCNB_Alloc_Pkt (pkt_len);
    if (!pkt)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (-1);
    }

    bzero (SMB_Hdr (pkt), SMB_ssetpLM_len);
    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 10;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = 0xFF; /* No extra command */
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, 0);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_mmc_offset, 2);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_vcn_offset, handle->pid);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_pwl_offset, strlen (passWd) + 1);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpLM_res_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char*) (SMB_Hdr (pkt) + SMB_ssetpLM_buf_offset);

    /* Copy  in password, then the rest. Password has a null at end
     */
    strcpy (p, passWd);
    p += strlen (passWd);

    strcpy (p, userName);
    p += strlen (userName) + 1;

    strcpy (p, handle->PDomain);
    p += strlen (handle->PDomain) + 1;

    strcpy (p, handle->OSName);
  }
  else
  {
    /* We don't admit to UNICODE support ...
     */
    param_len = strlen (userName) + 1 + strlen (passWd) +
                strlen (handle->PDomain) + 1 +
                strlen (handle->OSName) + 1 +
                strlen (handle->LMType) + 1;

    pkt_len = SMB_ssetpNTLM_len + param_len;
    pkt     = RFCNB_Alloc_Pkt (pkt_len);
    if (!pkt)
    {
      SMBlib_errno = SMBlibE_NoSpace;
      return (-1);
    }

    bzero (SMB_Hdr (pkt), SMB_ssetpNTLM_len);
    SIVAL (SMB_Hdr (pkt), SMB_hdr_idf_offset, SMB_DEF_IDF); /* Plunk in IDF */
    *(SMB_Hdr (pkt) + SMB_hdr_com_offset) = SMBsesssetupX;
    SSVAL (SMB_Hdr (pkt), SMB_hdr_pid_offset, handle->pid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_tid_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_mid_offset, handle->mid);
    SSVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset, handle->uid);
    *(SMB_Hdr (pkt) + SMB_hdr_wct_offset) = 13;
    *(SMB_Hdr (pkt) + SMB_hdr_axc_offset) = 0xFF; /* No extra command */
    SSVAL (SMB_Hdr (pkt), SMB_hdr_axo_offset, 0);

    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mbs_offset, SMBLIB_MAX_XMIT);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_mmc_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_vcn_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_snk_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cipl_offset, strlen (passWd));
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cspl_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_res_offset, 0);
    SIVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_cap_offset, 0);
    SSVAL (SMB_Hdr (pkt), SMB_ssetpNTLM_bcc_offset, param_len);

    /* Now copy the param strings in with the right stuff
     */
    p = (char*) (SMB_Hdr (pkt) + SMB_ssetpNTLM_buf_offset);

    /* Copy  in password, then the rest. Password has no null at end
     */
    strcpy (p, passWd);
    p += strlen (passWd);

    strcpy (p, userName);
    p += strlen (userName) + 1;

    strcpy (p, handle->PDomain);
    p += strlen (handle->PDomain) + 1;

    strcpy (p, handle->OSName);
    p += strlen (handle->OSName) + 1;

    strcpy (p, handle->LMType);
  }

  /* Now send it and get a response
   */
  if (RFCNB_Send (handle->transConnect, pkt, pkt_len) < 0)
  {
    SMB_debug (1, "Error sending SessSetupX request\n");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_SendFailed;
    return (-1);
  }

  /* Now get the response ...
   */
  if (RFCNB_Recv (handle->transConnect, pkt, pkt_len) < 0)
  {  
    SMB_debug  (1, "Error receiving response to SessSetupAndX\n");

    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_RecvFailed;
    return (-1);
  }

  /* Check out the response type ...
   */
  if (CVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset) != SMBC_SUCCESS)
  {  
    SMB_debug (1, "SMB_SessSetupAndX fail: errorclass = %i, Error Code = %i\n",
               CVAL (SMB_Hdr(pkt), SMB_hdr_rcls_offset),
               SVAL (SMB_Hdr(pkt), SMB_hdr_err_offset));

    SMBlib_SMB_Error = IVAL (SMB_Hdr (pkt), SMB_hdr_rcls_offset);
    RFCNB_Free_Pkt (pkt);
    SMBlib_errno = SMBlibE_Remote;
    return (-1);
  }

  SMB_debug (1, "SessSetupAndX response. Action = %i\n",
             SVAL (SMB_Hdr(pkt), SMB_ssetpr_act_offset));

  /* Now pick up the UID for future reference ...
   */
  handle->uid = SVAL (SMB_Hdr (pkt), SMB_hdr_uid_offset);
  RFCNB_Free_Pkt (pkt);
  return (0);
}


/*
 * Disconnect from the server, and disconnect all tree connects
 */
int SMB_Discon (struct SMB_Connect *handle, BOOL keepHandle)
{
  RFCNB_Hangup (handle->transConnect);

  if (!keepHandle)
     free (handle);
  return (0);
}
